﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.GeneralizedQuadraticAssignment.Common;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {

  [Item("RelocateEquipmentManipluator", "Relocates a random equipment from an overstuffed location to a random one with space or a random equipment if constraints are satisfied.")]
  [StorableClass]
  public class RelocateEquipmentManipluator : GQAPManipulator, IDemandsAwareGQAPOperator, ICapacitiesAwareGQAPOperator {

    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }

    [StorableConstructor]
    protected RelocateEquipmentManipluator(bool deserializing) : base(deserializing) { }
    protected RelocateEquipmentManipluator(RelocateEquipmentManipluator original, Cloner cloner) : base(original, cloner) { }
    public RelocateEquipmentManipluator()
      : base() {
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", GeneralizedQuadraticAssignmentProblem.DemandsDescription));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", GeneralizedQuadraticAssignmentProblem.CapacitiesDescription));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RelocateEquipmentManipluator(this, cloner);
    }

    public static void Apply(IRandom random, IntegerVector assignment, DoubleArray capacities, DoubleArray demands) {
      int locations = capacities.Length;
      if (locations < 2) throw new InvalidOperationException("There are not enough locations for relocation operations.");

      Dictionary<int, double> usedCapacities = new Dictionary<int, double>();
      Dictionary<int, List<int>> groupedLocations = new Dictionary<int, List<int>>();
      for (int i = 0; i < locations; i++) {
        usedCapacities[i] = 0.0;
        groupedLocations.Add(i, new List<int>());
      }
      for (int i = 0; i < assignment.Length; i++) {
        usedCapacities[assignment[i]] += demands[i];
        groupedLocations[assignment[i]].Add(i);
      }

      bool selected;
      var overstuffedLocation = usedCapacities.Where(x => capacities[x.Key] < x.Value).ChooseRandomOrDefault(random, out selected);
      if (selected) {
        int equipment = groupedLocations[overstuffedLocation.Key].ChooseRandomOrDefault(random, out selected);
        var freeLocations = usedCapacities.Where(x => capacities[x.Key] > x.Value).ToList();
        var location = freeLocations.Where(x => capacities[x.Key] > x.Value + demands[equipment]).ChooseRandomOrDefault(random, out selected);
        if (selected) {
          assignment[equipment] = location.Key;
        } else {
          location = freeLocations.ChooseRandomOrDefault(random, out selected);
          if (selected) assignment[equipment] = location.Key;
          else assignment[equipment] = random.Next(locations);
        }
      } else {
        assignment[random.Next(assignment.Length)] = random.Next(locations);
      }
    }

    protected override void Manipulate(IRandom random, IntegerVector vector) {
      Apply(random, vector, CapacitiesParameter.ActualValue, DemandsParameter.ActualValue);
    }
  }
}
